__title__ = "aiochlite"

__author__ = "darkstussy"

__copyright__ = f"Copyright (c) 2025 {__author__}"

from .client import AsyncChClient
from .core import ExternalTable, Row
from .exceptions import ChClientError

__all__ = ("AsyncChClient", "ChClientError", "ExternalTable", "Row")



from .client import ChClientCore, ClientCoreOptions
from .external_data import build_external_data
from .models import ExternalData, ExternalTable, Row

__all__ = (
    "ChClientCore",
    "ClientCoreOptions",
    "ExternalData",
    "ExternalTable",
    "Row",
    "build_external_data",
)



from http import HTTPStatus
from typing import Any, AsyncIterator, Mapping

from aiohttp import ClientResponse, ClientSession

from .exceptions import ChClientError


class HttpClient:
    """Wrapper around aiohttp ClientSession for HTTP operations."""

    def __init__(self, session: ClientSession):
        self._session = session

    async def get(self, url: str, params: Mapping[str, str]):
        async with self._session.get(url, params=params) as response:
            await _check_response(response)

    async def post(self, url: str, params: Mapping[str, str], *, data: Any = None) -> AsyncIterator[bytes] | None:
        async with self._session.post(url, params=params, data=data) as response:
            await _check_response(response)

    async def read(self, url: str, params: Mapping[str, str], *, data: Any = None) -> bytes:
        async with self._session.post(url, params=params, data=data) as response:
            await _check_response(response)
            return await response.read()

    async def stream(self, url: str, params: Mapping[str, str], *, data: Any = None) -> AsyncIterator[bytes]:
        async with self._session.post(url, params=params, data=data) as response:
            await _check_response(response)

            async for chunk in response.content.iter_chunked(262_144):
                yield chunk

    async def close(self):
        await self._session.close()


async def _check_response(response: ClientResponse):
    """Check HTTP response status and raise error if not OK."""
    if response.status != HTTPStatus.OK:
        raise ChClientError(await response.text(errors="replace"))



from datetime import date, datetime
from decimal import Decimal
from typing import Any
from uuid import UUID


class _MissingType:
    __slots__ = ()


_MISSING = _MissingType()


def _escape_ch_string_literal(value: str) -> str:
    return (
        value.replace("\\", "\\\\")
        .replace("\b", "\\b")
        .replace("\f", "\\f")
        .replace("\r", "\\r")
        .replace("\n", "\\n")
        .replace("\t", "\\t")
        .replace("\0", "\\0")
        .replace("'", "\\'")
    )


def _scalar_clickhouse_literal(value: Any) -> str | _MissingType:
    if value is None:
        out: str | _MissingType = "null"
    else:
        value_type = type(value)
        if value_type is bool:
            out = "true" if value else "false"
        elif value_type is int or value_type is float:
            out = str(value)
        elif value_type is str:
            out = f"'{_escape_ch_string_literal(value)}'"
        elif isinstance(value, datetime):
            out = f"'{value.strftime('%Y-%m-%d %H:%M:%S')}'"
        elif isinstance(value, date):
            out = f"'{value.strftime('%Y-%m-%d')}'"
        elif isinstance(value, (UUID, Decimal)):
            out = f"'{value}'"
        elif isinstance(value, bytes):
            out = f"'{_escape_ch_string_literal(value.decode('utf-8'))}'"
        else:
            out = _MISSING

    return out


def _container_clickhouse_literal(value: Any) -> str | _MissingType:
    if isinstance(value, tuple):
        return f"({','.join(_to_clickhouse_literal(item) for item in value)})"
    if isinstance(value, list):
        return f"[{','.join(_to_clickhouse_literal(item) for item in value)}]"
    if isinstance(value, dict):
        items = ",".join(f"{_to_clickhouse_literal(str(k))}:{_to_clickhouse_literal(v)}" for k, v in value.items())
        return f"{{{items}}}"

    return _MISSING


def _to_clickhouse_literal(value: Any) -> str:
    """Render Python value as a ClickHouse literal (used for container params)."""
    scalar = _scalar_clickhouse_literal(value)
    if not isinstance(scalar, _MissingType):
        return scalar

    container = _container_clickhouse_literal(value)
    if not isinstance(container, _MissingType):
        return container

    return f"'{_escape_ch_string_literal(str(value))}'"


def to_clickhouse(value: Any) -> str | int | float:
    """
    Convert Python value to ClickHouse parameter format.

    Args:
        value (Any): Python value to convert.

    Returns:
        str | int | float: Converted value suitable for ClickHouse.
    """
    if value is None:
        out: str | int | float = "NULL"
    else:
        value_type = type(value)
        if value_type is bool:
            out = 1 if value else 0
        elif value_type is int or value_type is float or value_type is str:
            out = value
        elif isinstance(value, (list, tuple, dict)):
            out = _to_clickhouse_literal(value)
        elif isinstance(value, datetime):
            out = value.strftime("%Y-%m-%d %H:%M:%S")
        elif isinstance(value, date):
            out = value.strftime("%Y-%m-%d")
        elif isinstance(value, (UUID, Decimal)):
            out = str(value)
        elif isinstance(value, bytes):
            out = value.decode("utf-8")
        else:
            out = str(value)

    return out



import ipaddress
import json
import re
import struct
from collections.abc import AsyncIterator, Sequence
from datetime import date, datetime, timedelta
from decimal import Decimal
from functools import lru_cache
from typing import Any, Callable, Iterable, Protocol
from uuid import UUID

from ._type_parsing import extract_base_type, extract_timezone, split_type_arguments, unwrap_wrappers


class _Reader(Protocol):
    def _read(self, size: int) -> memoryview: ...

    def read_int8(self) -> int: ...
    def read_int16(self) -> int: ...
    def read_int32(self) -> int: ...
    def read_int64(self) -> int: ...
    def read_uint8(self) -> int: ...
    def read_uint16(self) -> int: ...
    def read_uint32(self) -> int: ...
    def read_uint64(self) -> int: ...
    def read_float32(self) -> float: ...
    def read_float64(self) -> float: ...
    def read_varuint(self) -> int: ...
    def read_string(self) -> str: ...
    def skip(self, size: int) -> None: ...

    @property
    def pos(self) -> int: ...


class _BinaryReader:
    def __init__(self, data: bytes | memoryview):
        self._data = data if isinstance(data, memoryview) else memoryview(data)
        self._pos = 0

    def _read(self, size: int) -> memoryview:
        end = self._pos + size
        if end > len(self._data):
            raise ValueError("Unexpected end of data")
        chunk = self._data[self._pos : end]
        self._pos = end
        return chunk

    def read_uint8(self) -> int:
        if self._pos >= len(self._data):
            raise ValueError("Unexpected end of data")
        value = self._data[self._pos]
        self._pos += 1
        return int(value)

    def read_int8(self) -> int:
        if self._pos + 1 > len(self._data):
            raise ValueError("Unexpected end of data")
        value = struct.unpack_from("<b", self._data, self._pos)[0]
        self._pos += 1
        return value

    def read_uint16(self) -> int:
        value = struct.unpack_from("<H", self._data, self._pos)[0]
        self._pos += 2
        return value

    def read_int16(self) -> int:
        value = struct.unpack_from("<h", self._data, self._pos)[0]
        self._pos += 2
        return value

    def read_uint32(self) -> int:
        value = struct.unpack_from("<I", self._data, self._pos)[0]
        self._pos += 4
        return value

    def read_int32(self) -> int:
        value = struct.unpack_from("<i", self._data, self._pos)[0]
        self._pos += 4
        return value

    def read_uint64(self) -> int:
        value = struct.unpack_from("<Q", self._data, self._pos)[0]
        self._pos += 8
        return value

    def read_int64(self) -> int:
        value = struct.unpack_from("<q", self._data, self._pos)[0]
        self._pos += 8
        return value

    def read_int128(self) -> int:
        return int.from_bytes(self._read(16), "little", signed=True)

    def read_float32(self) -> float:
        value = struct.unpack_from("<f", self._data, self._pos)[0]
        self._pos += 4
        return value

    def read_float64(self) -> float:
        value = struct.unpack_from("<d", self._data, self._pos)[0]
        self._pos += 8
        return value

    def read_varuint(self) -> int:
        """Read LEB128 varuint."""
        shift = 0
        result = 0
        while True:
            byte = self.read_uint8()
            result |= (byte & 0x7F) << shift
            if byte < 0x80:
                break
            shift += 7
        return result

    def read_bytes(self, size: int) -> bytes:
        return self._read(size).tobytes()

    def read_string(self) -> str:
        length = self.read_varuint()
        return self._read(length).tobytes().decode("utf-8")

    @property
    def eof(self) -> bool:
        return self._pos >= len(self._data)

    @property
    def pos(self) -> int:
        return self._pos

    def skip(self, size: int) -> None:
        end = self._pos + size
        if end > len(self._data):
            raise ValueError("Unexpected end of data")
        self._pos = end


def _decimal_meta(ch_type: str) -> tuple[int, int]:
    inner = ch_type[ch_type.index("(") + 1 : ch_type.rindex(")")]
    parts = [p.strip() for p in inner.split(",")]

    if ch_type.startswith("Decimal32"):
        precision = 9
        scale = int(parts[0])
    elif ch_type.startswith("Decimal64"):
        precision = 18
        scale = int(parts[0])
    elif ch_type.startswith("Decimal128"):
        precision = 38
        scale = int(parts[0])
    elif ch_type.startswith("Decimal256"):
        precision = 76
        scale = int(parts[0])
    else:
        precision = int(parts[0])
        scale = int(parts[1])

    return precision, scale


def _decimal_size(precision: int) -> int:
    if precision <= 9:
        return 4
    if precision <= 18:
        return 8
    if precision <= 38:
        return 16
    if precision <= 76:
        return 32
    raise ValueError(f"Unsupported Decimal precision: {precision}")


_EPOCH_DATE = date(1970, 1, 1)


def _datetime_reader(ch_type: str) -> Callable[[_Reader], datetime]:
    tz = extract_timezone(ch_type)

    @lru_cache(maxsize=4096)
    def _dt(ts: int) -> datetime:
        return datetime.fromtimestamp(ts, tz=tz)

    def _read_dt(reader: _Reader) -> datetime:
        return _dt(reader.read_uint32())

    return _read_dt


def _datetime64_reader(ch_type: str) -> Callable[[_Reader], datetime]:
    inner = ch_type[ch_type.index("(") + 1 : ch_type.rindex(")")]
    parts = [p.strip() for p in inner.split(",")]
    scale = int(parts[0])
    tz = extract_timezone(ch_type)

    @lru_cache(maxsize=4096)
    def _dt64(ticks: int) -> datetime:
        base_seconds, remainder = divmod(ticks, 10**scale)
        dt = datetime.fromtimestamp(base_seconds, tz=tz)
        if remainder:
            micros = remainder * (10 ** (6 - scale)) if scale <= 6 else remainder / (10 ** (scale - 6))
            dt = dt + timedelta(microseconds=micros)
        return dt

    def _read_dt64(reader: _Reader) -> datetime:
        return _dt64(reader.read_int64())

    return _read_dt64


def _decimal_reader(ch_type: str) -> Callable[[_Reader], Decimal]:
    precision, scale = _decimal_meta(ch_type)
    size = _decimal_size(precision)

    @lru_cache(maxsize=4096)
    def _dec(raw: int) -> Decimal:
        return Decimal(raw).scaleb(-scale)

    def _read_dec(reader: _Reader) -> Decimal:
        return _dec(int.from_bytes(reader._read(size), "little", signed=True))

    return _read_dec


def _fixedstring_reader(ch_type: str) -> Callable[[_Reader], str]:
    inner = ch_type[ch_type.index("(") + 1 : ch_type.rindex(")")]
    size = int(inner.strip())

    def _read_fixedstring(reader: _Reader) -> str:
        raw = reader._read(size).tobytes()
        return raw.decode("utf-8", errors="replace").rstrip("\x00")

    return _read_fixedstring


@lru_cache(maxsize=512)
def _enum_mapping(ch_type: str) -> dict[int, str]:
    """
    Parse ClickHouse Enum8/Enum16 type definition into {value: name}.

    Example:
      Enum8('a' = 1, 'b' = 2)
    """
    inner = ch_type[ch_type.index("(") + 1 : ch_type.rindex(")")]
    pairs = re.findall(r"'((?:\\.|[^'])*)'\s*=\s*([+-]?\d+)", inner)
    if not pairs:
        raise ValueError(f"Invalid Enum definition: {ch_type}")

    mapping: dict[int, str] = {}
    for raw_name, raw_value in pairs:
        name = raw_name.replace("\\\\", "\\").replace("\\'", "'")
        mapping[int(raw_value)] = name

    return mapping


def _enum_reader(ch_type: str) -> Callable[[_Reader], str]:
    base = extract_base_type(ch_type)
    mapping = _enum_mapping(ch_type)

    if base == "Enum8":

        def _read_enum(reader: _Reader) -> str:
            value = reader.read_int8()
            return mapping.get(value, str(value))

        return _read_enum

    if base == "Enum16":

        def _read_enum(reader: _Reader) -> str:
            value = reader.read_int16()
            return mapping.get(value, str(value))

        return _read_enum

    raise ValueError(f"Unsupported Enum type: {ch_type}")


def _ipv4_reader(_: str) -> Callable[[_Reader], ipaddress.IPv4Address]:
    def _read_ipv4(reader: _Reader) -> ipaddress.IPv4Address:
        return ipaddress.IPv4Address(reader.read_uint32())

    return _read_ipv4


def _ipv6_reader(_: str) -> Callable[[_Reader], ipaddress.IPv6Address]:
    def _read_ipv6(reader: _Reader) -> ipaddress.IPv6Address:
        return ipaddress.IPv6Address(reader._read(16).tobytes())

    return _read_ipv6


def _array_reader(ch_type: str) -> Callable[[_Reader], list[Any]]:
    inner_type = ch_type[6:-1]
    inner = _reader_for_type(inner_type)

    def _read_array(reader: _Reader) -> list[Any]:
        return [inner(reader) for _ in range(reader.read_varuint())]

    return _read_array


def _map_reader(ch_type: str) -> Callable[[_Reader], dict[Any, Any]]:
    inner = ch_type[ch_type.index("(") + 1 : ch_type.rindex(")")]
    key_type, value_type = split_type_arguments(inner)
    key_reader = _reader_for_type(key_type)
    value_reader = _reader_for_type(value_type)

    def _read_map(reader: _Reader) -> dict[Any, Any]:
        count = reader.read_varuint()
        out: dict[Any, Any] = {}
        for _ in range(count):
            key = key_reader(reader)
            value = value_reader(reader)
            out[key] = value
        return out

    return _read_map


def _tuple_reader(ch_type: str) -> Callable[[_Reader], tuple[Any, ...]]:
    inner = ch_type[6:-1]
    element_types = split_type_arguments(inner)
    readers = tuple(_reader_for_type(t) for t in element_types)

    def _read_tuple(reader: _Reader) -> tuple[Any, ...]:
        return tuple(r(reader) for r in readers)

    return _read_tuple


def _uuid_reader(reader: _Reader) -> UUID:
    raw = reader._read(16).tobytes()
    return UUID(bytes=raw[:8][::-1] + raw[8:][::-1])


_PRIMITIVE_READERS: dict[str, Callable[[_Reader], Any]] = {
    "Bool": lambda r: r.read_uint8() != 0,
    "Float32": lambda r: r.read_float32(),
    "Float64": lambda r: r.read_float64(),
    "Int8": lambda r: r.read_int8(),
    "Int16": lambda r: r.read_int16(),
    "Int32": lambda r: r.read_int32(),
    "Int64": lambda r: r.read_int64(),
    "JSON": lambda r: json.loads(r.read_string()),
    "String": lambda r: r.read_string(),
    "UInt8": lambda r: r.read_uint8(),
    "UInt16": lambda r: r.read_uint16(),
    "UInt32": lambda r: r.read_uint32(),
    "UInt64": lambda r: r.read_uint64(),
}

_COMPLEX_READERS: dict[str, Callable[[str], Callable[[_Reader], Any]]] = {
    "Array": _array_reader,
    "Date": lambda _: (lambda r: _EPOCH_DATE + timedelta(days=r.read_uint16())),
    "Date32": lambda _: (lambda r: _EPOCH_DATE + timedelta(days=r.read_int32())),
    "DateTime": _datetime_reader,
    "DateTime64": _datetime64_reader,
    "Enum16": _enum_reader,
    "Enum8": _enum_reader,
    "FixedString": _fixedstring_reader,
    "IPv4": _ipv4_reader,
    "IPv6": _ipv6_reader,
    "Map": _map_reader,
    "Tuple": _tuple_reader,
    # ClickHouse UUID RowBinary is encoded as two UInt64 (hi, lo), each in little-endian.
    "UUID": lambda _: _uuid_reader,
}


@lru_cache(maxsize=256)
def _reader_for_type(ch_type: str) -> Callable[[_Reader], Any]:
    if ch_type.startswith("LowCardinality("):
        return _reader_for_type(ch_type[15:-1])

    if ch_type.startswith("Nullable("):
        inner = _reader_for_type(ch_type[9:-1])

        def _read_nullable(reader: _Reader) -> Any:
            return None if reader.read_uint8() else inner(reader)

        return _read_nullable

    ch_type = unwrap_wrappers(ch_type)
    base = extract_base_type(ch_type)

    primitive = _PRIMITIVE_READERS.get(base)
    if primitive is not None:
        return primitive

    if base.startswith("Decimal"):
        return _decimal_reader(ch_type)

    handler = _COMPLEX_READERS.get(base)
    if handler is not None:
        return handler(ch_type)

    raise ValueError(f"Unsupported RowBinary type: {ch_type}")


def parse_rowbinary_with_names_and_types(data: bytes) -> tuple[list[str], list[str], Iterable[list[Any]]]:
    """
    Parse RowBinaryWithNamesAndTypes payload and return header and row iterator.

    Returns:
        names: list of column names
        types: list of ClickHouse types
        rows: iterable of rows (list of values)
    """
    reader = _BinaryReader(data)
    column_count = reader.read_varuint()
    names = [reader.read_string() for _ in range(column_count)]
    types = [reader.read_string() for _ in range(column_count)]
    readers = [_reader_for_type(tp) for tp in types]

    def _rows() -> Iterable[list[Any]]:
        while not reader.eof:
            yield [read(reader) for read in readers]

    return names, types, _rows()


_FIXED_SIZES: dict[str, int] = {
    "Bool": 1,
    "UInt8": 1,
    "Int8": 1,
    "UInt16": 2,
    "Int16": 2,
    "UInt32": 4,
    "Int32": 4,
    "UInt64": 8,
    "Int64": 8,
    "Float32": 4,
    "Float64": 8,
    "Date": 2,
    "Date32": 4,
    "DateTime": 4,
    "Enum8": 1,
    "Enum16": 2,
    "IPv4": 4,
    "IPv6": 16,
}


def _array_skipper(inner_type: str) -> Callable[[_Reader], None]:
    inner_skip = _skipper_for_type(inner_type)

    inner_unwrapped = unwrap_wrappers(inner_type)
    inner_base = extract_base_type(inner_unwrapped)

    inner_fixed = _FIXED_SIZES.get(inner_base)
    if inner_fixed is not None:
        return lambda reader: reader.skip(reader.read_varuint() * inner_fixed)

    if inner_base == "DateTime64":
        return lambda reader: reader.skip(reader.read_varuint() * 8)

    if inner_base.startswith("Decimal"):
        precision, _scale = _decimal_meta(inner_unwrapped)
        size = _decimal_size(precision)
        return lambda reader: reader.skip(reader.read_varuint() * size)

    if inner_base == "UUID":
        return lambda reader: reader.skip(reader.read_varuint() * 16)

    def _skip_array(reader: _Reader) -> None:
        count = reader.read_varuint()
        for _ in range(count):
            inner_skip(reader)

    return _skip_array


def _map_skipper(ch_type: str) -> Callable[[_Reader], None]:
    inner = ch_type[ch_type.index("(") + 1 : ch_type.rindex(")")]
    key_type, value_type = split_type_arguments(inner)
    key_skip = _skipper_for_type(key_type)
    value_skip = _skipper_for_type(value_type)

    key_unwrapped = unwrap_wrappers(key_type)
    value_unwrapped = unwrap_wrappers(value_type)
    key_base = extract_base_type(key_unwrapped)
    value_base = extract_base_type(value_unwrapped)

    key_fixed = _FIXED_SIZES.get(key_base)
    value_fixed = _FIXED_SIZES.get(value_base)
    if key_fixed is not None and value_fixed is not None:
        pair_size = key_fixed + value_fixed
        return lambda reader: reader.skip(reader.read_varuint() * pair_size)

    if key_base == "UUID" and value_fixed is not None:
        return lambda reader: reader.skip(reader.read_varuint() * (16 + value_fixed))
    if value_base == "UUID" and key_fixed is not None:
        return lambda reader: reader.skip(reader.read_varuint() * (key_fixed + 16))

    def _skip_map(reader: _Reader) -> None:
        count = reader.read_varuint()
        for _ in range(count):
            key_skip(reader)
            value_skip(reader)

    return _skip_map


def _tuple_skipper(ch_type: str) -> Callable[[_Reader], None]:
    inner = ch_type[6:-1]
    element_types = split_type_arguments(inner)
    skippers = tuple(_skipper_for_type(t) for t in element_types)

    def _skip_tuple(reader: _Reader) -> None:
        for skip in skippers:
            skip(reader)

    return _skip_tuple


_COMPLEX_SKIPPERS: dict[str, Callable[[str], Callable[[_Reader], None]]] = {
    "Array": lambda ch_type: _array_skipper(ch_type[6:-1]),
    "DateTime64": lambda _: (lambda reader: reader.skip(8)),
    "JSON": lambda _: (lambda reader: reader.skip(reader.read_varuint())),
    "Map": _map_skipper,
    "String": lambda _: (lambda reader: reader.skip(reader.read_varuint())),
    "Tuple": _tuple_skipper,
    "UUID": lambda _: (lambda reader: reader.skip(16)),
}


@lru_cache(maxsize=256)
def _skipper_for_type(ch_type: str) -> Callable[[_Reader], None]:
    if ch_type.startswith("LowCardinality("):
        return _skipper_for_type(ch_type[15:-1])

    if ch_type.startswith("Nullable("):
        inner = _skipper_for_type(ch_type[9:-1])

        def _skip_nullable(reader: _Reader) -> None:
            if reader.read_uint8():
                return
            inner(reader)

        return _skip_nullable

    ch_type = unwrap_wrappers(ch_type)
    base = extract_base_type(ch_type)

    size = _FIXED_SIZES.get(base)
    if size is not None:
        return lambda reader: reader.skip(size)

    if base == "FixedString":
        inner = ch_type[ch_type.index("(") + 1 : ch_type.rindex(")")]
        fixed_size = int(inner.strip())
        return lambda reader: reader.skip(fixed_size)

    if base.startswith("Decimal"):
        precision, _scale = _decimal_meta(ch_type)
        size = _decimal_size(precision)
        return lambda reader: reader.skip(size)

    handler = _COMPLEX_SKIPPERS.get(base)
    if handler is not None:
        return handler(ch_type)

    raise ValueError(f"Unsupported RowBinary type: {ch_type}")


class RowBinaryLazyValues(Sequence[Any]):
    __slots__ = ("_cache", "_data", "_offsets", "_readers")
    _MISSING = object()

    def __init__(self, data: memoryview, offsets: list[tuple[int, int]], readers: list[Callable[[_Reader], Any]]):
        self._data = data
        self._offsets = offsets
        self._readers = readers
        self._cache: list[Any] = [self._MISSING] * len(offsets)

    def __len__(self) -> int:
        return len(self._offsets)

    def __getitem__(self, idx: int) -> Any:
        if idx < 0:
            idx += len(self._offsets)
        cached = self._cache[idx]
        if cached is not self._MISSING:
            return cached

        start, end = self._offsets[idx]
        reader = _BinaryReader(self._data[start:end])
        value = self._readers[idx](reader)
        self._cache[idx] = value
        return value


def parse_rowbinary_with_names_and_types_lazy(
    data: bytes,
) -> tuple[list[str], list[str], Iterable[RowBinaryLazyValues]]:
    """
    Parse RowBinaryWithNamesAndTypes payload and return rows with lazy per-cell decoding.
    """
    reader = _BinaryReader(data)
    column_count = reader.read_varuint()
    names = [reader.read_string() for _ in range(column_count)]
    types = [reader.read_string() for _ in range(column_count)]
    skippers = [_skipper_for_type(tp) for tp in types]
    readers = [_reader_for_type(tp) for tp in types]
    payload = memoryview(data)

    def _rows() -> Iterable[RowBinaryLazyValues]:
        while not reader.eof:
            offsets: list[tuple[int, int]] = []
            for skip in skippers:
                start = reader.pos
                skip(reader)
                end = reader.pos
                offsets.append((start, end))
            yield RowBinaryLazyValues(payload, offsets, readers)

    return names, types, _rows()


class _NeedMoreData(Exception):
    pass


class _StreamingReader:
    __slots__ = ("_buf", "_pos")

    def __init__(self):
        self._buf = bytearray()
        self._pos = 0

    def feed(self, data: bytes) -> None:
        if data:
            self._buf += data

    @property
    def pos(self) -> int:
        return self._pos

    @pos.setter
    def pos(self, value: int) -> None:
        self._pos = value

    @property
    def remaining(self) -> int:
        return len(self._buf) - self._pos

    @property
    def eof(self) -> bool:
        return self._pos >= len(self._buf)

    def compact(self) -> None:
        if self._pos and self._pos > 1_048_576:
            del self._buf[: self._pos]
            self._pos = 0

    def copy_slice(self, start: int, end: int) -> bytes:
        if end > len(self._buf):
            raise _NeedMoreData
        return bytes(self._buf[start:end])

    def _read(self, size: int) -> memoryview:
        end = self._pos + size
        if end > len(self._buf):
            raise _NeedMoreData
        mv = memoryview(self._buf)[self._pos : end]
        self._pos = end
        return mv

    def read_bytes(self, size: int) -> bytes:
        return self._read(size).tobytes()

    def skip(self, size: int) -> None:
        end = self._pos + size
        if end > len(self._buf):
            raise _NeedMoreData
        self._pos = end

    def read_uint8(self) -> int:
        if self._pos >= len(self._buf):
            raise _NeedMoreData
        b = self._buf[self._pos]
        self._pos += 1
        return b

    def read_int8(self) -> int:
        if self._pos + 1 > len(self._buf):
            raise _NeedMoreData
        value = struct.unpack_from("<b", self._buf, self._pos)[0]
        self._pos += 1
        return value

    def read_uint16(self) -> int:
        if self._pos + 2 > len(self._buf):
            raise _NeedMoreData
        value = struct.unpack_from("<H", self._buf, self._pos)[0]
        self._pos += 2
        return value

    def read_int16(self) -> int:
        if self._pos + 2 > len(self._buf):
            raise _NeedMoreData
        value = struct.unpack_from("<h", self._buf, self._pos)[0]
        self._pos += 2
        return value

    def read_uint32(self) -> int:
        if self._pos + 4 > len(self._buf):
            raise _NeedMoreData
        value = struct.unpack_from("<I", self._buf, self._pos)[0]
        self._pos += 4
        return value

    def read_int32(self) -> int:
        if self._pos + 4 > len(self._buf):
            raise _NeedMoreData
        value = struct.unpack_from("<i", self._buf, self._pos)[0]
        self._pos += 4
        return value

    def read_uint64(self) -> int:
        if self._pos + 8 > len(self._buf):
            raise _NeedMoreData
        value = struct.unpack_from("<Q", self._buf, self._pos)[0]
        self._pos += 8
        return value

    def read_int64(self) -> int:
        if self._pos + 8 > len(self._buf):
            raise _NeedMoreData
        value = struct.unpack_from("<q", self._buf, self._pos)[0]
        self._pos += 8
        return value

    def read_float32(self) -> float:
        if self._pos + 4 > len(self._buf):
            raise _NeedMoreData
        value = struct.unpack_from("<f", self._buf, self._pos)[0]
        self._pos += 4
        return value

    def read_float64(self) -> float:
        if self._pos + 8 > len(self._buf):
            raise _NeedMoreData
        value = struct.unpack_from("<d", self._buf, self._pos)[0]
        self._pos += 8
        return value

    def read_varuint(self) -> int:
        p = self._pos
        shift = 0
        result = 0
        while True:
            if p >= len(self._buf):
                raise _NeedMoreData
            byte = self._buf[p]
            p += 1
            result |= (byte & 0x7F) << shift
            if byte < 0x80:
                break
            shift += 7
        self._pos = p
        return result

    def read_string(self) -> str:
        p = self._pos
        length = self.read_varuint()
        if self._pos + length > len(self._buf):
            self._pos = p
            raise _NeedMoreData
        s = bytes(self._buf[self._pos : self._pos + length]).decode("utf-8")
        self._pos += length
        return s


class RowBinaryWithNamesAndTypesStreamParser:
    def __init__(self, chunks: AsyncIterator[bytes], *, lazy: bool = False):
        self._chunks = chunks.__aiter__()
        self._reader = _StreamingReader()
        self._done = False
        self._names: list[str] | None = None
        self._types: list[str] | None = None
        self._readers: list[Callable[[_Reader], Any]] | None = None
        self._skippers: list[Callable[[_Reader], None]] | None = None
        self._lazy = lazy

    async def _fill(self) -> bool:
        try:
            chunk = await anext(self._chunks)
        except StopAsyncIteration:
            self._done = True
            return False
        self._reader.feed(chunk)
        return True

    async def read_header(self) -> tuple[list[str], list[str]]:
        if self._names is not None and self._types is not None:
            return self._names, self._types

        while True:
            checkpoint = self._reader.pos
            try:
                column_count = self._reader.read_varuint()
                names = [self._reader.read_string() for _ in range(column_count)]
                types = [self._reader.read_string() for _ in range(column_count)]
                self._names = names
                self._types = types
                self._readers = [_reader_for_type(tp) for tp in types]
                if self._lazy:
                    self._skippers = [_skipper_for_type(tp) for tp in types]
                else:
                    self._skippers = None
            except _NeedMoreData:
                self._reader.pos = checkpoint
                if not await self._fill():
                    raise ValueError("Unexpected end of data") from None
            else:
                return names, types

    async def rows(self) -> AsyncIterator[list[Any] | RowBinaryLazyValues]:
        await self.read_header()
        assert self._readers is not None

        while True:
            if self._done and self._reader.remaining == 0:
                return

            checkpoint = self._reader.pos
            try:
                if self._lazy:
                    assert self._skippers is not None
                    row_start = self._reader.pos
                    offsets: list[tuple[int, int]] = []
                    for skip in self._skippers:
                        cell_start = self._reader.pos
                        skip(self._reader)
                        cell_end = self._reader.pos
                        offsets.append((cell_start - row_start, cell_end - row_start))
                    row_end = self._reader.pos
                    row_bytes = self._reader.copy_slice(row_start, row_end)
                    yield RowBinaryLazyValues(memoryview(row_bytes), offsets, self._readers)
                else:
                    values = [read(self._reader) for read in self._readers]
                    yield values
                self._reader.compact()
            except _NeedMoreData:
                self._reader.pos = checkpoint
                if not await self._fill():
                    if self._reader.remaining == 0:
                        return
                    raise ValueError("Unexpected end of data") from None



import re
from functools import lru_cache
from typing import Final
from zoneinfo import ZoneInfo

_TOP_LEVEL_COMMA_SPLIT_RE: Final[re.Pattern[str]] = re.compile(r",(?![^()]*\))")
_DATETIME_TZ_RE: Final[re.Pattern[str]] = re.compile(
    r"DateTime(?:64)?\(\s*(?:\d+\s*,\s*)?'([^']+)'\s*\)",
    re.IGNORECASE,
)


@lru_cache(maxsize=256)
def extract_base_type(ch_type: str) -> str:
    if ch_type.startswith("Nullable("):
        return extract_base_type(ch_type[9:-1])

    if ch_type.startswith("LowCardinality("):
        return extract_base_type(ch_type[15:-1])

    if "(" in ch_type:
        return ch_type[: ch_type.index("(")]

    return ch_type


@lru_cache(maxsize=256)
def unwrap_wrappers(ch_type: str) -> str:
    unwrapped = ch_type.strip()
    while True:
        if unwrapped.startswith("Nullable(") and unwrapped.endswith(")"):
            unwrapped = unwrapped[9:-1].strip()
            continue
        if unwrapped.startswith("LowCardinality(") and unwrapped.endswith(")"):
            unwrapped = unwrapped[15:-1].strip()
            continue
        return unwrapped


@lru_cache(maxsize=256)
def split_type_arguments(type_list: str) -> list[str]:
    return [part.strip() for part in _TOP_LEVEL_COMMA_SPLIT_RE.split(type_list) if part.strip()]


@lru_cache(maxsize=256)
def extract_timezone(ch_type: str) -> ZoneInfo | None:
    match = _DATETIME_TZ_RE.search(unwrap_wrappers(ch_type))
    if not match:
        return None

    tz = match.group(1)
    try:
        return ZoneInfo(tz)
    except Exception:
        return None



from collections.abc import Mapping
from typing import Any, TypedDict, Unpack

from aiochlite.converters import to_clickhouse

from .models import ExternalTable


class ClientCoreOptions(TypedDict, total=False):
    """Configuration options for ChClientCore."""

    user: str
    password: str
    database: str
    enable_compression: bool


class ChClientCore:
    """
    Core client logic for building requests and parsing responses.

    Args:
        user (str): ClickHouse username.
        password (str): ClickHouse password.
        database (str): Default database name.
        enable_compression (bool): Enable HTTP compression.
    """

    __slots__ = ("_database", "_enable_compression", "_password", "_user")

    def __init__(self, **kwargs: Unpack[ClientCoreOptions]):
        self._user = kwargs.get("user", "default")
        self._password = kwargs.get("password", "")
        self._database = kwargs.get("database", "default")
        self._enable_compression = kwargs.get("enable_compression", False)

    def build_headers(self) -> dict[str, str]:
        headers = {}
        if self._user:
            headers["X-ClickHouse-User"] = self._user
        if self._password:
            headers["X-ClickHouse-Key"] = self._password

        return headers

    def build_query_params(
        self,
        *,
        params: Mapping[str, Any] | None = None,
        settings: Mapping[str, Any] | None = None,
        external_tables: dict[str, ExternalTable] | None = None,
    ) -> dict[str, Any]:
        url_params: dict[str, Any] = {
            "database": self._database,
            "output_format_binary_write_json_as_string": 1,
        }

        if self._enable_compression:
            url_params["enable_http_compression"] = "1"

        if params:
            for key, value in params.items():
                url_params[f"param_{key}"] = to_clickhouse(value)

        if settings:
            url_params.update(settings)

        if external_tables:
            for name, external_table in external_tables.items():
                url_params[f"{name}_format"] = "JSONCompactEachRow"
                url_params[f"{name}_structure"] = ", ".join(
                    f"{column} {column_type}" for column, column_type in external_table.structure
                )

        return url_params



from collections.abc import AsyncIterator
from typing import Any, Mapping, Self, Sequence, TypedDict, Unpack

from aiohttp import ClientSession, FormData

from .converters import to_json
from .converters.rowbinary import (
    RowBinaryWithNamesAndTypesStreamParser,
    parse_rowbinary_with_names_and_types,
    parse_rowbinary_with_names_and_types_lazy,
)
from .core import ChClientCore, ClientCoreOptions, ExternalTable, Row, build_external_data
from .exceptions import ChClientError
from .http_client import HttpClient


class QueryOptions(TypedDict, total=False):
    """Options for ClickHouse query execution."""

    params: Mapping[str, Any] | None
    settings: Mapping[str, Any] | None
    external_tables: dict[str, ExternalTable] | None


class AsyncChClient:
    """
    Asynchronous ClickHouse HTTP client.

    Args:
        url (str): ClickHouse server URL.
        session (ClientSession | None): Optional aiohttp session to use.
        lazy_decode (bool): If True, decode row values lazily per cell (faster if you access only a subset of columns).
        user (str): ClickHouse username.
        password (str): ClickHouse password.
        database (str): Default database name.
        enable_compression (bool): Enable HTTP compression.
    """

    __slots__ = ("_core", "_database", "_http_client", "_lazy_decode", "_url")

    def __init__(
        self,
        url: str = "http://localhost:8123",
        *,
        session: ClientSession | None = None,
        lazy_decode: bool = True,
        **kwargs: Unpack[ClientCoreOptions],
    ):
        self._url = url
        self._database = kwargs.get("database", "default")
        self._lazy_decode = lazy_decode
        self._core = ChClientCore(**kwargs)

        headers = self._core.build_headers()
        session = session or ClientSession()
        session.headers.update(headers)
        self._http_client = HttpClient(session)

    async def __aenter__(self) -> Self:
        await self.ping(raise_on_error=True)
        return self

    async def __aexit__(self, *args):
        await self.close()

    async def close(self):
        """Close the underlying HTTP client session."""
        await self._http_client.close()

    async def ping(self, *, raise_on_error: bool = False) -> bool:
        """Check if ClickHouse server is reachable.

        Args:
            raise_on_error (bool): Whether to raise exception on connection failure.

        Returns:
            bool: True if server is alive, False otherwise.

        Raises:
            ChClientError: If raise_on_error is True and connection fails.
        """
        try:
            await self._http_client.get(self._url, params={**self._core.build_query_params(), "query": "SELECT 1"})
        except ChClientError:
            if raise_on_error:
                raise

            return False

        return True

    def _prepare_query(
        self,
        query: str,
        *,
        add_format: bool = True,
        **kwargs: Unpack[QueryOptions],
    ) -> tuple[dict[str, Any], str | FormData]:
        """Prepare query for execution by adding FORMAT clause (when needed) and building params."""
        if add_format:
            if "format" in query.lower():
                raise ValueError("The query must not contain a FORMAT clause.")

            query = f"{query} FORMAT RowBinaryWithNamesAndTypes"

        params = self._core.build_query_params(**kwargs)

        if external_tables := kwargs.get("external_tables"):
            data = FormData()
            for external_data in build_external_data(external_tables):
                data.add_field(
                    name=external_data.name,
                    value=external_data.content,
                    filename=external_data.filename,
                    content_type=external_data.content_type,
                )

            params["query"] = query
        else:
            data = query

        return params, data

    async def _stream(self, params: dict[str, Any], data: str | FormData) -> AsyncIterator[Row]:
        byte_chunks = self._http_client.stream(self._url, params=params, data=data)
        parser = RowBinaryWithNamesAndTypesStreamParser(byte_chunks, lazy=self._lazy_decode)
        names, _ = await parser.read_header()

        index = {name: idx for idx, name in enumerate(names)}
        async for values in parser.rows():
            yield Row(names, values, index=index)

    async def _fetch(self, params: dict[str, Any], data: str | FormData) -> list[Row]:
        payload = await self._http_client.read(self._url, params=params, data=data)
        names, _, rows = (
            parse_rowbinary_with_names_and_types_lazy(payload)
            if self._lazy_decode
            else parse_rowbinary_with_names_and_types(payload)
        )

        index = {name: idx for idx, name in enumerate(names)}
        return [Row(names, values, index=index) for values in rows]

    async def execute(self, query: str, **kwargs: Unpack[QueryOptions]):
        """Execute query without returning results.

        Raises:
            ChClientError: If query execution fails.
        """
        params, data = self._prepare_query(query, add_format=False, **kwargs)
        await self._http_client.post(self._url, params=params, data=data)

    async def stream(self, query: str, **kwargs: Unpack[QueryOptions]) -> AsyncIterator[Row]:
        """Execute query and iterate over results.

        Yields:
            Row: Query result rows.

        Raises:
            ChClientError: If query execution fails.
        """
        params, data = self._prepare_query(query, **kwargs)
        async for row in self._stream(params, data):
            yield row

    async def fetch(self, query: str, **kwargs: Unpack[QueryOptions]) -> list[Row]:
        """Execute query and fetch all results.

        Returns:
            list[Row]: List of all result rows.

        Raises:
            ChClientError: If query execution fails.
        """
        params, data = self._prepare_query(query, **kwargs)
        return await self._fetch(params, data)

    async def fetchone(self, query: str, **kwargs: Unpack[QueryOptions]) -> Row | None:
        """Execute query and fetch first result row.

        Returns:
            Row | None: First row or None if no results.

        Raises:
            ChClientError: If query execution fails.
        """
        async for row in self.stream(query, **kwargs):
            return row

        return None

    async def fetchval(self, query: str, **kwargs: Unpack[QueryOptions]) -> Any:
        """Execute query and fetch first column of first row.

        Returns:
            Any: First column value or None if no results.

        Raises:
            ChClientError: If query execution fails.
        """
        if row := await self.fetchone(query, **kwargs):
            return row.first()

        return None

    async def insert(
        self,
        table: str,
        data: Sequence[dict[str, Any]] | list[tuple[Any, ...]],
        *,
        database: str | None = None,
        column_names: Sequence[str] | None = None,
        settings: Mapping[str, Any] | None = None,
    ):
        """Insert data into a ClickHouse table.

        Args:
            table (str): Table name.
            data (Sequence[dict[str, Any]] | list[tuple[Any, ...]]): Rows to insert.
            database (str | None): Database name (uses default if None).
            column_names (Sequence[str] | None): Column names for tuple data.
            settings (Mapping[str, Any] | None): ClickHouse settings.

        Raises:
            ChClientError: If insertion fails.
        """
        if not data:
            return

        db = database or self._database

        columns_clause = f" ({', '.join(column_names)})" if column_names else ""

        if isinstance(data[0], dict):
            format_name = "JSONEachRow"
            body = "\n".join(to_json(row) for row in data)
        else:
            format_name = "JSONCompactEachRow"
            body = "\n".join(to_json(list(row)) for row in data)

        await self._http_client.post(
            self._url,
            params=self._core.build_query_params(settings=settings),
            data=f"INSERT INTO {db}.{table}{columns_clause} FORMAT {format_name}\n{body}",
        )



import json
from typing import Any, Sequence

from .models import ExternalData, ExternalTable


def _transform_to_json_compact_each_row(columns: list[str], rows: Sequence[Any]) -> list[tuple[Any, ...]]:
    """Transform dict rows to tuple rows in column order."""
    return [tuple(row[col] for col in columns) for row in rows]


def _to_json_compact_each_row_bytes(external_table: ExternalTable) -> bytes:
    """Convert external table to JSONCompactEachRow format bytes."""
    rows = external_table.data
    if isinstance(rows[0], dict):
        rows = _transform_to_json_compact_each_row([s[0] for s in external_table.structure], rows)

    data = "\n".join(json.dumps(list(r), ensure_ascii=False, separators=(",", ":")) for r in rows) + "\n"
    return data.encode("utf-8")


def build_external_data(external_tables: dict[str, ExternalTable]) -> list[ExternalData]:
    """Build external data objects for multipart request."""
    return [
        ExternalData(
            name,
            content=_to_json_compact_each_row_bytes(external_table),
            filename=f"{name}.json",
            content_type="application/json",
        )
        for name, external_table in external_tables.items()
    ]



import json
from datetime import date, datetime
from decimal import Decimal
from typing import Any
from uuid import UUID


def _json_default(value: Any) -> str:
    if isinstance(value, datetime):
        return value.strftime("%Y-%m-%d %H:%M:%S")
    if isinstance(value, date):
        return value.strftime("%Y-%m-%d")
    if isinstance(value, (UUID, Decimal)):
        return str(value)
    if isinstance(value, bytes):
        return value.decode("utf-8")
    return str(value)


def to_json(data: Any) -> str:
    """Convert Python data to JSON string for ClickHouse HTTP API."""
    return json.dumps(data, default=_json_default, ensure_ascii=False, separators=(",", ":"))



from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from typing import Any, Iterator, NamedTuple


@dataclass(slots=True)
class ExternalTable:
    """External table data for ClickHouse queries.

    Attributes:
        structure (Sequence[tuple[str, str]]): Column definitions as (name, type) pairs.
        data (Sequence[dict[str, Any]] | Sequence[tuple[Any, ...]]): Table rows as dicts or tuples.
    """

    structure: Sequence[tuple[str, str]]
    data: Sequence[dict[str, Any]] | Sequence[tuple[Any, ...]]


class ExternalData(NamedTuple):
    """External data file representation for multipart requests."""

    name: str
    content: bytes
    filename: str
    content_type: str | None = None


class Row:
    """Query result row with column access by name or index."""

    __slots__ = ("_dict", "_index", "_names", "_values")

    def __init__(self, names: list[str], values: Sequence[Any], *, index: Mapping[str, int] | None = None):
        self._names = names
        self._values = values
        self._index = index
        self._dict: dict[str, Any] | None = None

    def _as_dict(self) -> dict[str, Any]:
        if self._dict is None:
            self._dict = dict(zip(self._names, self._values, strict=False))
        return self._dict

    def __getattr__(self, name: str) -> Any:
        if self._index is not None and name in self._index:
            return self._values[self._index[name]]

        mapping = self._as_dict()
        if name in mapping:
            return mapping[name]

        raise AttributeError(f"Row has no column '{name}'")

    def __getitem__(self, key: str) -> Any:
        if self._index is not None and key in self._index:
            return self._values[self._index[key]]
        return self._as_dict()[key]

    def __iter__(self) -> Iterator[Any]:
        return iter(self._names)

    def __len__(self) -> int:
        return len(self._names)

    def __repr__(self) -> str:
        return f"Row({self._as_dict()})"

    def first(self) -> Any:
        """Get value of the first column."""
        return self._values[0] if self._values else None
